import inquirer, { Question } from 'inquirer';
import { readFileSync, writeFileSync } from 'fs';
import replace from 'replace-in-file';
import path from 'path';
import shelljs from 'shelljs';
import chalk from 'chalk';
const { mv, rm, which, exec } = shelljs;
const { green, cyan, red, yellow, underline } = chalk;

export default function initialize(
  dirname: string,
  modifyFiles = [
    // defaults
    'index.html',
    'README.md',
    'package.json',
    'vcf-element.ts',
    'demo/demo.js',
    'demo/index.html',
    'src/vcf-element.ts',
    'src/mixins/CustomEventMixin.ts',
    'test/vcf-element.test.ts',
    'util/publish.js'
  ],
  renameFiles = [
    // defaults
    'vcf-element.ts',
    'src/vcf-element.ts',
    'test/vcf-element.test.ts'
  ]
): void {
  const rmDirs = ['.git', '.vscode'];
  const rmFiles = ['util/init.js'];

  /* UTIL */

  const required = (input?: string) => {
    const valid = Boolean(input);
    if (!valid) throw Error('This is required');
    return valid;
  };

  const error = () => {
    console.log(red('There was an error building the workspace'));
    process.exit(1);
  };

  /* QUESTIONS */

  const questionElementName: Question = {
    type: 'input',
    name: 'element',
    message: cyan('What do you want the element to be called? (use kebab-case)'),
    validate: input => {
      if (required(input) && /^[a-z][\da-z]*(-[\da-z]+)*$/g.test(input)) return true;
      throw Error('"kebab-case" uses lowercase letters, and hyphens for any punctuation');
    }
  };

  const questionElementDescription: Question = {
    type: 'input',
    name: 'description',
    message: cyan('Please provide a description of the element')
  };

  const questionElementSuggest: Question = {
    type: 'confirm',
    name: 'useSuggestedName',
    message: cyan(`Would you like the web component to be called "${elementNameSuggested()}"?`),
    validate: input => required(input)
  };

  const questionRemoveGit: Question = {
    type: 'confirm',
    name: 'removegit',
    message: cyan('Would you like to clear the git history?'),
    validate: input => required(input)
  };

  /* MAIN */

  function main() {
    console.clear();

    if (!which('git')) {
      console.log(red('Sorry, this script requires git'));
      removeItems();
      process.exit(1);
    }

    // Greeting
    console.log(`Vaadin Component Factory ${cyan('}>')}\n`);

    // Generate the element name and start the tasks
    if (!elementNameSuggestedIsDefault()) {
      elementNameSuggestedAccept();
    } else {
      elementNameCreate();
    }
  }

  /**
   * Asks the user for the name of the element if it has been cloned into the
   * default directory, or if they want a different name to the one suggested
   */
  function elementNameCreate() {
    inquirer
      .prompt([questionElementName])
      .then(answers => elementDescriptionCreate(answers.element))
      .catch(error);
  }

  /**
   * Asks the user for a decription of the element to be used in package.json and summary for jsdocs
   */
  function elementDescriptionCreate(elementname: string) {
    inquirer
      .prompt([questionElementDescription])
      .then(answers => removeGitChoice(elementname, answers.description))
      .catch(error);
  }

  /**
   * Asks the user whether they would like to reset git history
   */
  function removeGitChoice(elementname: string, elementdescription: string) {
    inquirer
      .prompt([questionRemoveGit])
      .then(answers => setupElement(elementname, elementdescription, answers.removegit))
      .catch(error);
  }

  /**
   * Sees if the users wants to accept the suggested element name if the project
   * has been cloned into a custom directory (i.e. it's not 'vcf-element')
   */
  function elementNameSuggestedAccept() {
    inquirer
      .prompt([questionElementSuggest])
      .then(answers => {
        if (answers.useSuggestedName) {
          elementDescriptionCreate(elementNameSuggested());
        } else {
          elementNameCreate();
        }
      })
      .catch(error);
  }

  /**
   * The element name is suggested by looking at the directory name of the
   * tools parent directory and converting it to kebab-case
   *
   * The regex for this looks for any non-word or non-digit character, or
   * an underscore (as it's a word character), and replaces it with a dash.
   * Any leading or trailing dashes are then removed, before the string is
   * lowercased and returned
   */
  function elementNameSuggested() {
    return path
      .basename(path.resolve(dirname, '..'))
      .replace(/\W|_/g, '-')
      .replace(/^-+|-+$/g, '')
      .toLowerCase();
  }

  /**
   * Checks if the suggested element name is the default, which is 'vcf-element'
   */
  function elementNameSuggestedIsDefault() {
    return elementNameSuggested() === 'vcf-element';
  }

  /**
   * Calls all of the functions needed to setup the element
   */
  function setupElement(elementname: string, elementdescription: string, removegit = false) {
    console.log(cyan('\nThanks for the info. The last few changes are being made...\n\n'));

    // Get the Git username and email before the .git directory is removed
    const username = exec('git config user.name').stdout.trim();
    const usermail = exec('git config user.email').stdout.trim();

    removeItems(removegit);

    modifyContents(elementname, elementdescription, username, usermail);

    renameItems(elementname);

    finalize(removegit);

    console.log(cyan("You're all set. Happy coding! :)\n"));
  }

  /**
   * Removes items from the project that aren't needed after the initial setup
   */
  function removeItems(removegit = false) {
    console.log(underline.white('Removed'));

    // The directories and files are combined here, to simplify the function,
    // as the 'rm' command checks the item type before attempting to remove it
    let rmItems = rmFiles;
    if (removegit) {
      rmItems = [...rmDirs, ...rmFiles];
    }

    rm(
      '-rf',
      rmItems.map(f => path.resolve(dirname, '..', f))
    );

    console.log(red(rmItems.join('\n')));

    console.log('\n');
  }

  /**
   * Updates the contents of the template files with the element name or user details
   */
  function modifyContents(elementname: string, elementdescription: string, username: string, usermail: string) {
    console.log(underline.white('Modified'));

    const elementclassname = elementname
      .split('-')
      .map(c => c[0].toUpperCase() + c.slice(1))
      .join('');

    const files = modifyFiles.map(f => path.resolve(dirname, '..', f));
    try {
      replace.sync({
        files,
        from: [
          /vcf-element(?!-util)/g,
          /VcfElement/g,
          /--elementdescription--/g,
          /--username--/g,
          /--usermail--/g,
          /\*\*Note.*\*\*/,
          /\*\*Run.*\*\*/
        ],
        to: [elementname, elementclassname, elementdescription, username, usermail, '', '']
      });
      console.log(yellow(modifyFiles.join('\n')));
    } catch (error) {
      console.error('An error occurred modifying the file:', error);
    }

    console.log('\n');
  }

  /**
   * Renames any template files to the new element name
   */
  function renameItems(elementname: string) {
    console.log(underline.white('Renamed'));

    for (const file of renameFiles) {
      const newFilename = file.replace(/vcf-element/g, elementname);
      mv(path.resolve(dirname, '..', file), path.resolve(dirname, '..', newFilename));
      console.log(cyan(file + ' => ' + newFilename));
    }

    console.log('\n');
  }

  /**
   * Calls any external programs to finish setting up the element
   */
  function finalize(removegit = false) {
    console.log(underline.white('Finalizing'));

    // Recreate Git folder
    if (removegit) {
      const gitInitOutput = exec('git init "' + path.resolve(dirname, '..') + '"', {
        silent: true
      }).stdout;
      console.log(green(gitInitOutput.replace(/(\n|\r)+/g, '')));
    }

    // Remove post-install command
    const jsonPackage = path.resolve(dirname, '..', 'package.json');
    const pkg = JSON.parse(readFileSync(jsonPackage).toString());

    // Note: Add items to remove from the package file here
    delete pkg.scripts.postinstall;

    writeFileSync(jsonPackage, JSON.stringify(pkg, null, 2));
    console.log(green('Postinstall script has been removed'));

    console.log('\n');
  }

  /* RUN */

  main();
}
